/*************************************************************/
/*  A model of phyllotactic patterning in small meristems    */
/*   Authors: Przemyslaw Prusinkiewicz and Mikolaj Cieslak   */
/*             Copyrights (c) 2024 by the authors            */
/*       Distributed under the CC BY-NC-SA 4.0 licence       */
/*************************************************************/

// Switch between parameters for Figure 4ab and 4cd in
// E. Coen and P. Prusinkiewicz "Developmental timing in Plants",
// Nature Communications 15:2674 (2024).

#include <lpfgall.h>

#define RadToDeg 57.29577951
#define TwoPi 6.283185307

/* Simulation parameters */
// #define Fig4ab 	// comment out to get Fig. 4cd

#ifdef Fig4ab
	#define rho0 2.00	// initial radius of primordia
	#define meristem_growth_duration 0
	#define active_ring_expansion_duration 0
	#define show_parastichies 0
#else
	#define rho0 1.22	// initial radius of primordia
	#define meristem_growth_duration 0
	#define active_ring_expansion_duration 2000
	#define show_parastichies 0 // set to 1 to visualize (3,5) parastichies
#endif	

#define s0 		1.0		// initial arc-length position of the active ring
#define dms		0.000	// meristem size increment per simulation step
#define drho	0.0003	// primordium size increment per simulation step
#define ds		0.005	// primordium displacement (s coordinate ncrement per simulation step)

#define dphi	0.01	// angular step for scanning active ring, in radians
#define nds 	0.0001	// normalized arc length used for integrating the generating curve length

#define max(a,b)	(((a) >= (b)) ? (a) : (b))
#define min(a,b)	(((a) >= (b)) ? (b) : (a))
#define distance(a,b)	((a)-(b)).Length()

float CURVE_SCALE;	// scaling factor for the graphically-defined apex profile
float L;	 	 	// length of the generating curve

/* Calculate position of a point on the scaled generative curve 
   as a function of the normalized arc-length coordinate ns
   (ns = 0: top of the apex; ns = 1: base of the apex). */

V3f profileXYZ(float ns)
{
	return CURVE_SCALE * curveXYZf(MeristemContour,ns);
} 

/* Generate curve normal at position ns. */
V3f profileNormal(float ns)
{
	return curveNormal(MeristemContour,ns);
}

/* Compute current length of the scaled generating curve
   (recalculated in each step "from scratch" in case a more
   dynamic profile is used in a more complex version of the program). */
float CurveLength()
{
	float AccumulatedLength = 0;
	for (float ns=0; ns<1.0; ns+=nds)
  		AccumulatedLength +=
			 (profileXYZ(ns+nds) - profileXYZ(ns)).Length();
	return AccumulatedLength;
}

/* Compute meristem radius at the curvilinear coordinate s.  
   s is the actual (not normalized) arc-length position on the
   profile curve (arc-length distance from the apex tip). */
float Radius(float s) 
{
	return (profileXYZ(s/L)).x;
}

/* Compute elevation corresponding to curvilinear coordinate s.
   Apex tip (s==0) has elevation > 0, apex base (s==L) has elevation 0, 
   stem (s>L) (assumed to be cylindrical) has elevations L-s < 0. */
float Elevation(float s) 
{
	return (profileXYZ(s/L)).y + ((s<L) ? 0 : L-s) ;
}

/* Data structure defining a primordium */
struct primordium{
	double rho;	// primordium radius (assumed constant while patterning)
	double s;	// arc-length position along the generating curve
	double phi;	// angular cylindrical coordinate
	double h;	// vertical cylindrical coordinate
	double r;	// horizontal cylindrical coordinate
	V3d pos;	// Cartesian position
	float age;	// age, needed to visualize primordium growth
	double insertion_time; // needed for the azimuth vs. time plot
};

primordium P[10000];	// array of primordia

primordium tp;	// trial primordium on the active ring

/* Coordinate system conversion and synchronization: 
   update cylindrical and Cartesian coordinates, given p.s and p.phi. */
void SyncCoordinates(primordium *p)
{
	p->h = Elevation (p->s);	 // convert to cylindrical coordinates
	p->r = Radius (p->s);

	p->pos.x = p->r*cos(p->phi); // convert to Cartesian coordinates
	p->pos.z = p->r*sin(p->phi);
	p->pos.y = p->h;
	
	return;
}

/* Check if tp collides with previous primordia. For optimization purposes,
   only consider 60 last primordia. */
int HofmeisterTest(primordium tp)
{
	extern int n;
	for (int i = max(0, n-60); i<n; i++) {
		if (distance(tp.pos, P[i].pos) < tp.rho+P[i].rho) // if not enough space...
			return 1;
	}
	return 0;
}
				
int n;	// number of primordia created so far
float stem_length;
float CurrentTime;

module Draw(); 				// draw the 3D pattern of primordia (+ the stem) and the plot
module MeristemSectors();	// draw meristem as a set of vertical sectors
module MeristemSlices();	// draw meristem incl. central zone as a sequence of horizontal slices
module MeristemMesh();		// draw meristem incl. central zone as a 3D mesh
module ActiveRingSpheres();	// draw active ring as a set of spheres
module ActiveRingGenCyl();	// draw active ring as a generalized cylinder 

derivation length: 1;

/* Alternative axioms test different representations of the meristem. */
// Axiom: SB MeristemSectors() EB SB ActiveRingSpheres() EB SB Draw() EB ;
// Axiom: SB MeristemSectors() EB SB ActiveRingGenCyl() EB SB Draw() EB ;
// Axiom: SB MeristemSlices() EB SB Draw() EB ;
Axiom: SB MeristemMesh() EB SB Draw() EB;

/* Initialize everything */
Start:
{
	CurrentTime = 0;

	CURVE_SCALE = 3.5;
	L = CurveLength() ;
	
	tp.rho = rho0;
	tp.s = s0;
	n = 0;				// there is no initial primordium
	tp.age = 0;
	tp.insertion_time = 0;

	UseView(Pattern); 	// visualize the 3D pattern
  	UseView(Plot);		// create the azimuth vs. time plot
}

/* Main simulation loop */
StartEach: {
	CurrentTime += 1;

	// Iterate over azimuths of the tentative primordium tp, insert a new 
	// primordium at tp if tp does not collide with those previously formed. 
	for (tp.phi = 0; tp.phi < TwoPi; tp.phi += dphi) {
		SyncCoordinates (&tp);
		if (HofmeisterTest(tp) == 0) {
			P[n] = tp;
			P[n].insertion_time=CurrentTime;
			n++;
		}
	}

	// Simulate uniform meristem expansion.
	if (CurrentTime < meristem_growth_duration)
		CURVE_SCALE += dms;

	// Calculate/update the generative curve length and the resulting growth ratio.
	float oldL = L;
	L = CurveLength() ;
	float ratio = L / oldL;

	// Advance arc-length position of the active ring.
	if (CurrentTime < active_ring_expansion_duration) {
		if (tp.s < oldL)	// update arc-length coordinate to preserve relative position...
			tp.s *= ratio;	// ...(normalized arc-length coordinate) in the meristem
		else 				// if active ring is in the stem - should never happen
			tp.s += oldL * (ratio - 1);
		tp.s += drho;		// simulate expansion with respect to the meristem
	}
	
	// Advance arc-length position and age of primordia - similar to active ring (tp).
	for (int i=0; i<n; i++) {
		if(P[i].s < oldL) 		// update arc-length coordinate to keep relative position
			P[i].s *= ratio;	// (normalized arc-length coordinate) within the meristem
		else
			P[i].s += oldL * (ratio - 1); 	// update arc-length position within stem to
		P[i].s += ds;						// account for the increase of s at the meristem base 
		P[i].age += 1;
		SyncCoordinates(&P[i]);
	}

	// Elongate the stem so that all primordia fit.
	stem_length = max(0, -P[0].h+P[0].rho);
}

/* Visualize simulation results in two view areas (subwindows): as a 3D 
   representation of the developing structure, and as a 2D plot showing
   the emergence of primordia over time */
interpretation:

/* 3D visualization in the top view area (subwindow). 
   Different meristem and active ring visualization alternatives
   and can be chosen be selecting the appropriate axiom. */
vgroup Pattern:

// Visualize active ring as a set of spheres (most straightforward).
ActiveRingSpheres(): {
	nproduce SB f(stem_length+tp.h) SetColor (11);
	float ang_incr = 3.0;
	for (float ang = 0; ang < 360; ang += ang_incr) {
           nproduce SB RollR(ang) Right(90) f(tp.r) Sphere(0.15) EB;
	}
	produce EB;
}

// Visualize active ring as a generalized cylinder (fewer polygons).
ActiveRingGenCyl(): {
	float n = 50.0;
	float l = 2.0*tp.r*sin(TwoPi/(2*n));
	nproduce SB f(stem_length+tp.h) Down(90) f(tp.r) Left(90)
		CurrentContour(0) SetWidth(0.30) SetColor(11)
		Left(180/n) StartGC();	
	for(float i=0; i<n; i++)
		nproduce Left(180/n) F(l) Left(180/n);
	produce F(0)EndGC() EB;
}

// Visualize meristem as a surface of revolution, made of vertical sectors. 
// Straightforward, but the central zone cannot be visualized this way.
MeristemSectors(): {
	nproduce SB f(stem_length) Down(90) SetColor(15) CurrentContour(1) 
		ScaleContour(2*CURVE_SCALE, 2*CURVE_SCALE) StartGC;
        float ang_incr = 5.0;
	for (float ang = 0; ang < 360; ang += ang_incr) 
           nproduce Right(ang_incr) F(0.0);
	produce EndGC EB;
}

// Visualize meristem as a generalized cylinder made of horizontal slices
// The central zone is visualized, but shading is off because lpfg calculates 
// normal vectors incorrectly as if all slices were cylindrical,
// ignoring that in general they are truncated cones. 
MeristemSlices() : {
	float radius = 2*profileXYZ(1).x;
	float height;
	float dns = 0.01;
	nproduce SB f(stem_length) ScaleContour(radius,radius) StartGC;
	for (float ns=1.0; ns>0.5*dns; ns-=dns) {
		radius = 2*profileXYZ(ns-dns).x;
		height = profileXYZ(ns-dns).y - profileXYZ(ns).y;
		nproduce SetColor((ns<0.7*tp.s/L) ? 12 : 15);
		nproduce ScaleContour(radius, radius) F(height); 
	}
	produce EndGC EB;
} 

// Visualize meristem as a surface of revolution made of polygons defined 
// explicitly using SP ... EP modules (i.e., a polygon mesh, without
// calls to StartGC / EndGC). Correct, including properly visualized
// central zone, but the code is long.
MeristemMesh() : {

  float dy = stem_length;

  const int LONGITUDINAL_SECS = 80;

  nproduce SB;

  for (int j = 0; j < LONGITUDINAL_SECS; ++j) {
    double phi1 = double(j) / double(LONGITUDINAL_SECS) * 2. * M_PI;
    double phi2 = double(j+1) / double(LONGITUDINAL_SECS) * 2. * M_PI;

    V3d p1, n1, t1;
    V3d p2, n2, t2;
    V3d p3, n3, t3;
    V3d p4, n4, t4;

    // For the first transverse section, compute the bottom two points:
    double s1 = 1.;
   
    // first pt, normal and tangent at this value of 's' on the curve,
    V3d pc1 = profileXYZ(s1);
    p1.Set(pc1.x*cos(phi1), pc1.y, -pc1.x*sin(phi1));
    V3d nc1 = profileNormal(s1);
    n1.Set(nc1.x*cos(phi1), nc1.y, -nc1.x*sin(phi1));
    n1 = n1.Normalize();
    t1.Set(-nc1.y*cos(phi1), nc1.x, nc1.y*sin(phi1));
    t1 = t1.Normalize();
  
    // next point, same s but next phi.
    V3d pc4 = pc1;
    p4.Set(pc4.x*cos(phi2), pc4.y, -pc4.x*sin(phi2));
    V3d nc4 = nc1;
    n4.Set(nc4.x*cos(phi2), nc4.y, -nc4.x*sin(phi2));
    n4 = n4.Normalize();
    t4.Set(-nc4.y*cos(phi2), nc4.x, nc4.y*sin(phi2));
    t4 = t4.Normalize();


    for (float s2 = 0.01; s2<1; s2 += 0.01) {

      V3d pc2 = profileXYZ(s2);
      p2.Set(pc2.x*cos(phi1), pc2.y, -pc2.x*sin(phi1));
      V3d nc2 = profileNormal(s2);
      n2.Set(nc2.x*cos(phi1), nc2.y, -nc2.x*sin(phi1));
      n2 = n2.Normalize();
      t2.Set(-nc2.y*cos(phi1), nc2.x, nc2.y*sin(phi1));
      t2 = t2.Normalize();
  
      V3d pc3 = pc2;
      p3.Set(pc3.x*cos(phi2), pc3.y, -pc3.x*sin(phi2));
      V3d nc3 = nc2;
      n3.Set(nc3.x*cos(phi2), nc3.y, -nc3.x*sin(phi2));
      n3 = n3.Normalize();
      t3.Set(-nc3.y*cos(phi2), nc3.x, nc3.y*sin(phi2));
      t3 = t3.Normalize();

	  int colour;

	  if (s2<0.7*tp.s/L)
      	colour = 13;
	  else
      	colour = 14;

      nproduce SetColor(colour);float ang_incr = 5.0;

      // Draw a quad (normal vector set using turtle 'up' vector, 
	  // but make sure to set 'heading' to tangent).
      nproduce	SP
  		MoveTo(p1.x,p1.y+dy,p1.z) SetHead(t1.x,t1.y,t1.z,n1.x,n1.y,n1.z) PP
  		MoveTo(p2.x,p2.y+dy,p2.z) SetHead(t2.x,t2.y,t2.z,n2.x,n2.y,n2.z) PP
  		MoveTo(p3.x,p3.y+dy,p3.z) SetHead(t3.x,t3.y,t3.z,n3.x,n3.y,n3.z) PP
  		MoveTo(p4.x,p4.y+dy,p4.z) SetHead(t4.x,t4.y,t4.z,n4.x,n4.y,n4.z) PP 
    		EP;

      // Copy the top points of this quad to the bottom points of the next quad.
      p1 = p2;
      n1 = n2;
      t1 = t2;
      p4 = p3;
      n4 = n3;
      t4 = t3;
    }

    // Draw a triangle as the last element (get apex point at tip [s==1, phi==0]).
    pc1 = profileXYZ(0.);
    p1.Set(pc1.x, pc1.y, 0.);
    nc1 = profileNormal(0.);
    n1.Set(nc1.x, nc1.y, 0.);
    n1 = n1.Normalize();
    t1.Set(-nc1.y, nc1.x, 0.);
    t1 = t1.Normalize();

    nproduce SP
 		    MoveTo(p1.x,p1.y+dy,p1.z) SetHead(t1.x,t1.y,t1.z,n1.x,n1.y,n1.z) PP
  		    MoveTo(p3.x,p3.y+dy,p3.z) SetHead(t3.x,t3.y,t3.z,n3.x,n3.y,n3.z) PP
  		    MoveTo(p2.x,p2.y+dy,p2.z) SetHead(t2.x,t2.y,t2.z,n2.x,n2.y,n2.z) PP
		    EP;

  }
  nproduce EB;
  produce;
}

// Draw stem and primordia. 
Draw():
{
	// Draw stem
	nproduce SB CurrentContour(0) SetWidth(2*Radius(L)) MoveTo(0,0,0) SetColor(15) 
		LineTo(0,stem_length,0) EB ; 

	float ang_incr = 5.0;

	// Visualize primordia.
	int color;
	V3d primordium_position;
	for (int i=0; i<n; i++) {
		color = 1;

	// Optionally highlight (3,5) parastichies.
	if(show_parastichies) {
		if (i>1) {
			if (i % 3 == 0)
				color = 2;
			if (i % 5 == 0)
				color = 3;
			if ((i % 3 == 0) && (i % 5 == 0))
				color = 4;
		}
	}

		primordium_position = P[i].pos;
		// Move all primordia vertically by the length of the stem.
		// Primordia in the meristem have positive y coordinates
		// and thus appear above the top of the stem.
		// Primordia in the stem have negative y coordinates
		// and thus appear below the top of the stem.
		// Stem length is calculated so that the lowest primordium
		// P[0] appears at the bottom of the stem. 

		// Uncomment to see primordium coordinates
		// Printf("Primordium %d: s = %f, y = %f\n", i, P[i].s, P[i].pos.y);

		primordium_position.y += stem_length;	// ...so that the plant grows up
		nproduce SetColor(color) MoveTo3d(primordium_position) 
		Sphere(rho0*1.05*func(PrimordiumGrowth, 0.001* P[i].age));
	}
	produce;
}

/* Create the dynamic 2D azimuth vs. time plot showing primordia emergence. */
vgroup Plot:

Draw:
{
	// Draw and label the x axis.
	nproduce CurrentContour(0);
	nproduce SB MoveTo(0,0,0) SetColor(7) SetWidth(0.02) Right(90) F(7.0) 
		SB SetWidth(0.1) StartGC() SetWidth(0.0) F(0.1) EndGC() EB 
		f(-0.18) Left (90) f(0.1) SetColor(7) Label("Time") EB;

	// Draw and label the y axis.
	nproduce SB MoveTo(0,0,0) SetColor(7) SetWidth (0.02) F(1.0) 
		SB SetWidth(0.1) StartGC() SetWidth(0.0) F(0.1) EndGC() EB 
		f(-0.0) Right (90) f(0.1) SetColor(6) Label("Azimuth [deg]") EB;

	// Draw and label ticks on the y axis.
	nproduce SB SetColor(7) SetWidth(0.02) MoveTo(0,0,0) LineTo (-0.05, 0,0) 
		MoveTo(-0.18, -0.03,0) SetColor(6) Label("0") EB ;
	nproduce SB SetColor(7) SetWidth(0.02) MoveTo(0,0.45,0) LineTo (-0.05, 0.45,0) 
		MoveTo(-0.33, 0.42,0) SetColor(6) Label("180") EB ;
	nproduce SB SetColor(7) SetWidth(0.02) MoveTo(0,0.9,0) LineTo (-0.05, 0.9,0) 
		MoveTo(-0.33, 0.87,0) SetColor(6) Label("360") EB ;

	// Draw horizontal reference line indicating azimuth = 360 degrees.
	nproduce SB SetColor(9) MoveTo(0,0.9,0) Right(90) SetWidth(0.010) F(7.0) EB ;

	// Draw a moving vertical line indicating current time. 
	nproduce SB SetColor(9) MoveTo(0.0012*CurrentTime,0,0) SetWidth(0.015) F(0.9) EB ;

	// Draw primordia positions.
	nproduce SetColor(7);
	for (int i=0; i<n; i++) {
		nproduce MoveTo(0.0012*P[i].insertion_time, 0.9*(P[i].phi - 0*M_PI) / TwoPi, 0) Circle(0.05); 
	}

	produce;
}
